Package: brassica
Type: Package
Title: 1970s BASIC Interpreter
Version: 1.0.3
Date: 2026-04-15
Authors@R: person(given = "Mike", family = "Lee", role = c("aut", "cre"),
                  email = "random.deviate@gmail.com")
Author: Mike Lee [aut, cre]
Maintainer: Mike Lee <random.deviate@gmail.com>
Description: Executes BASIC programs from the 1970s, for historical and
    educational purposes. This enables famous examples of early machine
    learning, artificial intelligence, natural language processing, cellular
    automata, and so on, to be run in their original form.
License: GPL-3
Depends: R (>= 3.3.0)
NeedsCompilation: no
Packaged: 2026-04-15 03:14:46 UTC; Mike
Repository: CRAN
Date/Publication: 2026-04-15 06:00:02 UTC
Built: R 4.6.0; ; 2026-04-26 13:33:02 UTC; unix
